/*
 * Copyright (C) 2018 Google LLC. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#import "config.h"
#import "SafeBrowsingController.h"

#if WK_API_ENABLED

#import "APINavigationAction.h"
#import "SafeBrowsingContextProvider.h"
#import "SafeBrowsingSPI.h"
#import "WKNavigationActionInternal.h"
#import "WKPreferences.h"
#import "WKWebViewConfiguration.h"
#import "WKWebViewInternal.h"
#import "WebFramePolicyListenerProxy.h"
#import "WebPageProxy.h"

namespace WebKit {

const char* SafeBrowsingController::s_warningPageUrl = "file:///SafeBrowsing";

SafeBrowsingController::SafeBrowsingController(WKWebView *webView)
    : m_webView(webView)
{
}

SafeBrowsingController::~SafeBrowsingController()
{
}

void SafeBrowsingController::checkURL(Ref<API::NavigationAction>&& navigationAction, NavigationPolicyDecider&& policyDecider, Ref<WebFramePolicyListenerProxy>&& listener)
{
    bool shouldSkipSafeBrowsingLookUp = !m_webView.configuration.preferences.safeBrowsingEnabled;
    if (!shouldSkipSafeBrowsingLookUp) {
        bool warningBypassed = navigationAction->navigationType() == WebCore::NavigationType::LinkClicked
            && navigationAction->sourceFrame()
            && navigationAction->sourceFrame()->page()->pageLoadState().url() == s_warningPageUrl
            && navigationAction->sourceFrame()->page()->currentURL() == navigationAction->originalURL().string();

        if (warningBypassed)
            m_warningIgnoredUrls.add(navigationAction->originalURL());

        shouldSkipSafeBrowsingLookUp = warningBypassed || m_warningIgnoredUrls.contains(navigationAction->originalURL());
    }

    auto safeBrowsingCheck = std::make_unique<SafeBrowsingCheck>(WTFMove(listener), WTFMove(navigationAction));
    auto checkRaw = safeBrowsingCheck.get();
    m_checks.add(WTFMove(safeBrowsingCheck));

    if (shouldSkipSafeBrowsingLookUp)
        receivedSafeBrowsingResult(checkRaw, SafeBrowsingResult::Safe);
    else {
#if (PLATFORM(MAC) && __MAC_OS_X_VERSION_MIN_REQUIRED >= 101300) || (PLATFORM(IOS) && __IPHONE_OS_VERSION_MIN_REQUIRED >= 110000)
        auto weakThis = m_weakPtrFactory.createWeakPtr(*this);
        [SafeBrowsingContextProvider::sharedLookupContext() lookUpURL:checkRaw->navigationAction->request().url() completionHandler:^(SSBLookupResult* result, NSError* error) {
            if (!weakThis)
                return;

            SafeBrowsingResult safeBrowsingResult = SafeBrowsingResult::Safe;
            if (result.serviceLookupResults) {
                for (SSBServiceLookupResult* details in result.serviceLookupResults) {
                    // FIXME: Keep track of the different types in order to give different warnings.
                    if (details.isPhishing || details.isMalware || details.isUnwantedSoftware || details.isKnownToBeUnsafe)
                        safeBrowsingResult = SafeBrowsingResult::Unsafe;
                }
            }
            weakThis->receivedSafeBrowsingResult(checkRaw, safeBrowsingResult);
        }];
#else
        receivedSafeBrowsingResult(checkRaw, SafeBrowsingResult::Safe);
#endif
    }

    auto policyCompletionHandler = [checkRaw, weakThis = m_weakPtrFactory.createWeakPtr(*this)] (WKNavigationActionPolicy policy, std::optional<WebsitePoliciesData>&& data) mutable {
        if (weakThis)
            weakThis->receivedNavigationPolicyDecision(*checkRaw, policy, WTFMove(data));
    };

    policyDecider(WTFMove(policyCompletionHandler));
}

void SafeBrowsingController::receivedNavigationPolicyDecision(SafeBrowsingController::SafeBrowsingCheck& check, WKNavigationActionPolicy policy, std::optional<WebsitePoliciesData>&& data)
{
    check.navigationPolicyDecision = policy;
    check.websitePoliciesData = WTFMove(data);
    maybeFinishedCheck(check);
}

void SafeBrowsingController::receivedSafeBrowsingResult(SafeBrowsingCheck* check, SafeBrowsingResult safeBrowsingResult)
{
    if (!m_checks.contains(check))
        return;

    check->safeBrowsingResult = safeBrowsingResult;
    maybeFinishedCheck(*check);
}

void SafeBrowsingController::maybeFinishedCheck(SafeBrowsingController::SafeBrowsingCheck& check)
{
    if (!check.navigationPolicyDecision)
        return;

    if (check.navigationPolicyDecision.value() == WKNavigationActionPolicyAllow && !check.safeBrowsingResult)
        return;

    switch (check.navigationPolicyDecision.value()) {
    case WKNavigationActionPolicyAllow: {
        if (check.safeBrowsingResult == SafeBrowsingResult::Safe) {
            check.listener->use(WTFMove(check.websitePoliciesData));
            break;
        }

        check.listener->ignore();
        // FIXME: This really needs to be loaded in a way that permits localization.
        // FIXME: Make the warning message more similar to Safari, which means different
        // wording depending on the type of warning (phishing, malware, etc.).
        auto targetMainFrameUrl = check.navigationAction->targetFrame()->isMainFrame() ? check.navigationAction->request().url() : WebCore::URL(WebCore::ParsedURLString, check.navigationAction->targetFrame()->page()->currentURL());
        NSString* warningString = [NSString stringWithFormat:@"<style>body { background-color: red; }</style><body><h1>This is a simple SafeBrowsing warning.</h1><h2>You've been warned!!!</h2><h2><a id='visitAnyway' href='%s'>Visit the site anyway</a></h2>", targetMainFrameUrl.string().utf8().data()];
        [m_webView _loadAlternateHTMLString:warningString baseURL:WebCore::URL(WebCore::ParsedURLString, s_warningPageUrl) forUnreachableURL:targetMainFrameUrl];
        break;
    }
    case WKNavigationActionPolicyCancel:
        check.listener->ignore();
        break;
    case _WKNavigationActionPolicyDownload:
        check.listener->download();
        break;
    default:
        ASSERT_NOT_REACHED();
    }
    m_checks.remove(&check);
}

} // namespace WebKit

#endif // WK_API_ENABLED
