/*
 * Copyright (C) 2018 Google LLC. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. AND ITS CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL APPLE INC. OR ITS CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#if PLATFORM(MAC) && __MAC_OS_X_VERSION_MIN_REQUIRED >= 101300

#import "PlatformUtilities.h"
#import "TestProtocol.h"
#import <WebKit/WKNavigationActionPrivate.h>
#import <WebKit/WKWebViewConfiguration.h>
#import <WebKit/WKWebViewPrivate.h>
#import <wtf/RetainPtr.h>

#if WK_API_ENABLED

static bool shouldCancelNavigation;
static bool decidedNavigationPolicy;
static bool finishedNavigation;
static bool finishedRedirectNavigation;

static NSString *firstURL = @"sb://host/good.html";
static NSString *secondURL = @"sb://host/good2.html";
static NSString *unsafeURL = @"sb://host/phishing.html";
static NSString *warningURL = @"file:///SafeBrowsing";

@interface SBScheme : NSObject <WKURLSchemeHandler>
@end

@implementation SBScheme

- (void)webView:(WKWebView *)webView startURLSchemeTask:(id <WKURLSchemeTask>)urlSchemeTask
{
    NSString *data = @"Lorem ipsum dolor sit amet";
    [urlSchemeTask didReceiveResponse:[[[NSURLResponse alloc] initWithURL:urlSchemeTask.request.URL MIMEType:@"text/html" expectedContentLength:data.length textEncodingName:nil] autorelease]];
    [urlSchemeTask didReceiveData:[data dataUsingEncoding:NSUTF8StringEncoding]];
    [urlSchemeTask didFinish];
}

- (void)webView:(WKWebView *)webView stopURLSchemeTask:(id <WKURLSchemeTask>)urlSchemeTask
{
}

@end

@interface SafeBrowsingTestController : NSObject <WKNavigationDelegate, WKUIDelegate>
@end

@implementation SafeBrowsingTestController

- (void)webView:(WKWebView *)webView decidePolicyForNavigationAction:(WKNavigationAction *)navigationAction decisionHandler:(void (^)(WKNavigationActionPolicy))decisionHandler
{
    if (shouldCancelNavigation)
        decisionHandler(WKNavigationActionPolicyCancel);
    else
        decisionHandler(WKNavigationActionPolicyAllow);
    decidedNavigationPolicy = true;
}

- (void)webView:(WKWebView *)webView didFinishNavigation:(WKNavigation *)navigation
{
    if (finishedNavigation)
        finishedRedirectNavigation = true;
    else
        finishedNavigation = true;
}


@end

TEST(WebKit, SafeBrowsingForSafePage)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], "");
}

TEST(WebKit, SafeBrowsingForUnsafePage)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:unsafeURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[[[webView backForwardList] backItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], unsafeURL);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView evaluateJavaScript:@"var evt = document.createEvent('MouseEvent'); evt.initMouseEvent('click'); document.getElementById('visitAnyway').dispatchEvent(evt);" completionHandler:nil];
    TestWebKitAPI::Util::run(&finishedNavigation);
    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[[[webView backForwardList] backItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], "");
}


TEST(WebKit, SafeBrowsingForUnsafePageBackForward)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:unsafeURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[[[webView backForwardList] backItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], unsafeURL);

    finishedNavigation = false;
    [webView goBack];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[[[webView backForwardList] forwardItem] URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], firstURL);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView goForward];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[[[webView backForwardList] backItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], unsafeURL);

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:secondURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    [webView goBack];
    TestWebKitAPI::Util::run(&finishedNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    // FIXME: Also check backItem and forwardItem, once those are updated correctly.

    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], unsafeURL);
}


TEST(WebKit, SafeBrowsingForUnsafePageCancelNavigation)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    decidedNavigationPolicy = false;
    shouldCancelNavigation = true;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:unsafeURL]]];
    TestWebKitAPI::Util::run(&decidedNavigationPolicy);
    shouldCancelNavigation = false;

    EXPECT_FALSE(finishedNavigation);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView URL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], firstURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], "");
}

TEST(WebKit, SafeBrowsingForUnsafePageRedirect)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    finishedRedirectNavigation = false;
    decidedNavigationPolicy = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[[NSBundle mainBundle] URLForResource:@"redirect-to-unsafe" withExtension:@"html" subdirectory:@"TestWebKitAPI.resources"]]];
    TestWebKitAPI::Util::run(&finishedRedirectNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], unsafeURL);
    // FIXME: Also check that backItem is firstURL, once that backItem is updated correctly.

    EXPECT_WK_STREQ([[webView URL] absoluteString], unsafeURL);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], unsafeURL);
}

TEST(WebKit, SafeBrowsingForUnsafeSubframe)
{
    RetainPtr<WKWebViewConfiguration> configuration = adoptNS([[WKWebViewConfiguration alloc] init]);
    RetainPtr<SBScheme> handler = adoptNS([[SBScheme alloc] init]);
    [configuration setURLSchemeHandler:handler.get() forURLScheme:@"SB"];
    [configuration preferences].safeBrowsingEnabled = YES;
    auto webView = adoptNS([[WKWebView alloc] initWithFrame:NSMakeRect(0, 0, 800, 600) configuration:configuration.get()]);
    [webView _setUseTestSafeBrowsingContext:YES];

    auto window = adoptNS([[NSWindow alloc] initWithContentRect:[webView frame] styleMask:NSWindowStyleMaskBorderless backing:NSBackingStoreBuffered defer:YES]);
    [[window contentView] addSubview:webView.get()];

    auto controller = adoptNS([[SafeBrowsingTestController alloc] init]);
    [webView setNavigationDelegate:controller.get()];
    [webView setUIDelegate:controller.get()];

    finishedNavigation = false;
    [webView loadRequest:[NSURLRequest requestWithURL:[NSURL URLWithString:firstURL]]];
    TestWebKitAPI::Util::run(&finishedNavigation);

    finishedNavigation = false;
    finishedRedirectNavigation = false;
    decidedNavigationPolicy = false;
    NSURLRequest* request = [NSURLRequest requestWithURL:[[NSBundle mainBundle] URLForResource:@"unsafe-iframe" withExtension:@"html" subdirectory:@"TestWebKitAPI.resources"]];
    [webView loadRequest:request];
    TestWebKitAPI::Util::run(&finishedRedirectNavigation);

    EXPECT_TRUE(decidedNavigationPolicy);
    EXPECT_WK_STREQ([[[[webView backForwardList] currentItem] URL] absoluteString], [[request URL] absoluteString]);
    // FIXME: Also check that backItem is firstURL, once that backItem is updated correctly.

    EXPECT_WK_STREQ([[webView URL] absoluteString], [[request URL] absoluteString]);
    EXPECT_WK_STREQ([[webView _committedURL] absoluteString], warningURL);
    EXPECT_WK_STREQ([[webView _unreachableURL] absoluteString], [[request URL] absoluteString]);
}

#endif

#endif
